"""Class to load pre-built data set information from JSON files in source .  All results are cached"""
import os.path as osp
import json
from lrgasp import LrgaspException
from lrgasp.objDict import ObjDict
from lrgasp.defs import Species, Sample, LibraryPrep, Platform

##
# LRGASP run metadata files in lib/lrgasp/data/
##
lrgasp_run_metadata_files = ("encode-metadata.json", "manatee-metadata.json")


class LrgaspRun(ObjDict):
    """Used to create an LRGASP run object for sequencing data. Class is only uses
    to serialized, ObjDict used to access when deserialized."""
    def __init__(self, *, species, sample, run_acc, description, library_prep, platform):
        self.species = species
        self.sample = sample
        self.run_acc = run_acc
        self.description = description
        self.library_prep = library_prep
        self.platform = platform
        self.replicates = None

class LrgaspReplicate(ObjDict):
    """collection of files for a replicate. Class is only uses
    to serialized, ObjDict used to access when deserialized."""
    def __init__(self, replicate_number):
        self.replicate_number = replicate_number
        self.files = []

class LrgaspRnaSeqFile(ObjDict):
    """One data file from an RNA-Seq run, uses to build JSON in a consistency way
    Class is only uses to serialized, ObjDict used to access when
    deserialized."""
    def __init__(self, *, file_acc, file_type, url, s3_uri, file_size, md5sum, run_acc, biological_replicate_number,
                 output_type, paired_end=None, paired_with=None):
        self.file_acc = file_acc
        self.file_type = file_type
        self.url = url
        self.s3_uri = s3_uri
        self.file_size = file_size
        self.md5sum = md5sum
        self.run_acc = run_acc
        self.biological_replicate_number = biological_replicate_number
        self.output_type = output_type
        self.paired_end = paired_end
        self.paired_with = paired_with

class LrgaspRnaSeqMetaData(list):
    """deserialized LRSGAP RNA-Seq metadata, along with access methods"""
    cache = None

    def __init__(self):
        self.by_file_acc = {}
        self.by_run_acc = {}

    def _edit_run_types(self, run):
        "conversion to SymEnum"
        run.species = Species(run.species)
        run.sample = Sample(run.sample)
        run.library_prep = LibraryPrep(run.library_prep)
        run.platform = Platform(run.platform)

    def _add_files(self, files):
        for fil in files:
            if fil.file_acc in self.by_file_acc:
                raise LrgaspException("duplicate file accession: " + fil.file_acc)
            self.by_file_acc[fil.file_acc] = fil

    def add(self, run):
        self._edit_run_types(run)
        if run.run_acc in self.by_run_acc:
            raise LrgaspException("duplicate run id: " + run.run_acc)
        self.append(run)
        self.by_run_acc[run.run_acc] = run
        for replicate in run.replicates:
            self._add_files(replicate.files)

    def get_file_by_acc(self, file_acc):
        try:
            return self.by_file_acc[file_acc]
        except KeyError:
            raise LrgaspException(f"unknown LRGASP file accession {file_acc}, file accession should be from the LRGASP RNA-Seq data matrix")

    def get_run_by_acc(self, run_acc):
        try:
            return self.by_run_acc[run_acc]
        except KeyError:
            raise LrgaspException(f"unknown LRGASP run accession {run_acc}, run accession should be from the LRGASP RNA-Seq data matrix")

    def get_run_by_file_acc(self, file_acc):
        fil = self.get_file_by_acc(file_acc)
        return self.get_run_by_acc(fil.run_acc)

def _load_lrgasp_rna_seq_metadata_file(rna_seq_md, metadata_json):
    with open(metadata_json) as fh:
        for run in json.load(fh, object_pairs_hook=ObjDict):
            rna_seq_md.add(run)

def _load_lrgasp_rna_seq_metadata_files():
    "load of all metadata files when not cached"
    rna_seq_md = LrgaspRnaSeqMetaData()
    for json_file in lrgasp_run_metadata_files:
        json_path = osp.join(osp.dirname(__file__), "data", json_file)
        try:
            _load_lrgasp_rna_seq_metadata_file(rna_seq_md, json_path)
        except Exception as ex:
            raise LrgaspException(f"failed to load {json_path}") from ex
    return rna_seq_md

def get_lrgasp_rna_seq_metadata():
    """get LRGASP metadata, possible cached"""
    if LrgaspRnaSeqMetaData.cache is None:
        LrgaspRnaSeqMetaData.cache = _load_lrgasp_rna_seq_metadata_files()
    return LrgaspRnaSeqMetaData.cache



"""
Definitions of metadata identifiers, types, and functions to operate on them.
"""
import re
from collections import defaultdict
from lrgasp import LrgaspException
from lrgasp.symEnum import SymEnum, auto

# fixed file names, without .gz
ENTRY_JSON = "entry.json"
EXPERIMENT_JSON = "experiment.json"
MODELS_GTF = "models.gtf"
READ_MODEL_MAP_TSV = "read_model_map.tsv"
EXPRESSION_TSV = "expression.tsv"

class Challenge(SymEnum):
    """Challenge identifiers, value matches challenge number"""
    iso_detect_ref = 1
    iso_quant = 2
    iso_detect_de_novo = 3

class ExperimentType(SymEnum):
    "type of a experiment, assumed from Challenge type"
    model = auto()
    expression = auto()

class DataCategory(SymEnum):
    """categories of experiments based on data accepted"""
    long_only = auto()
    short_only = auto()
    long_short = auto()
    kitchen_sink = auto()

class Platform(SymEnum):
    """Simplified sequencing platform, mostly used if figuring out LibraryCategory"""
    Illumina = auto()
    PacBio = auto()
    ONT = auto()

class LibraryPrep(SymEnum):
    """Type of library prep"""
    CapTrap = auto()
    dRNA = auto()
    R2C2 = auto()
    cDNA = auto()

class Sample(SymEnum):
    """LRGASP sample identifier"""
    WTC11 = auto()
    H1_mix = auto()
    ES = auto()
    Manatee = auto()

class Species(SymEnum):
    """Species identifiers"""
    human = auto()
    mouse = auto()
    manatee = auto()
    simulated = auto()

class Repository(SymEnum):
    """Public data repositories"""
    SRA = auto()
    ENA = auto()
    INSDC = auto()
    ENC = auto()

class RefGenome(SymEnum):
    """LRGASP reference genomes"""
    GRCh38 = auto()
    GRCm39 = auto()

class Gencode(SymEnum):
    """LRGASP GENCODE version"""
    GENCODE_V38 = auto()
    GENCODE_VM27 = auto()

class ExpressionUnits(SymEnum):
    "Units used in expression matrix"
    RPM = auto()
    RPKM = auto()
    FPKM = auto()
    TPM = auto()
    counts = auto()

def validate_symbolic_ident(ident):
    if not ident.isidentifier():
        raise LrgaspException(f"not a valid symbolic identifier: '{ident}'")

def validate_feature_ident(ident):
    if (not ident.isascii()) or (not ident.isprintable()) or re.search("\\s", ident) or (len(ident) == 0):
        raise LrgaspException(f"invalid feature identifier, must be composed of ASCII, printable, non-white-space characters: '{ident}'")

def validate_synapse_ident(ident):
    if not re.match("^syn[0-9]{4,30}$", ident):
        raise LrgaspException(f"not a valid Synapse identifier: '{ident}'")

def validate_entry_ident(entry_id):
    """check that an entry is prefix with one of the challenge ids, return the Challenge identifier that
    matches"""
    for ch in Challenge:
        if entry_id.startswith(str(ch) + '_'):
            return ch
    valid_pre = ", ".join([str(ch) + '_*' for ch in Challenge])
    raise LrgaspException("entry_id {} must be prefixed with a challenge id ({})".format(entry_id, valid_pre))

def challenge_to_experiment_type(challenge):
    assert isinstance(challenge, Challenge)
    if challenge is Challenge.iso_quant:
        return ExperimentType.expression
    else:
        return ExperimentType.model

def sample_to_species(sample):
    if sample in (Sample.WTC11, Sample.H1_mix):
        return Species.human
    elif sample == Sample.ES:
        return Species.model
    elif sample == Sample.Manatee:
        return Species.manatee
    else:
        raise LrgaspException(f"bug mapping sample to species: {sample}")


_challenge_sample_map = {
    Challenge.iso_detect_ref: frozenset([Sample.WTC11, Sample.H1_mix, Sample.ES]),
    Challenge.iso_quant: frozenset([Sample.WTC11, Sample.H1_mix]),
    Challenge.iso_detect_de_novo: frozenset([Sample.Manatee, Sample.ES]),
}

def _build_sample_challenge_map():
    # reverse mapping
    s2c = defaultdict(set)
    for c in _challenge_sample_map:
        for s in _challenge_sample_map[c]:
            s2c[s].add(c)
    s2c.default_factory = None
    for s in s2c.keys():
        s2c[s] = tuple(sorted(s2c[s]))
    return s2c

# challenges are sorted by value
_sample_challenge_map = _build_sample_challenge_map()

def get_challenge_samples(challenge_id):
    return _challenge_sample_map[challenge_id]

def sample_to_challenges(sample):
    return _sample_challenge_map[sample]



"""
Simplified GTF parser/validator
"""
import numpy as np
from gtfparse import read_gtf, ParsingError
from lrgasp import LrgaspException

class Transcript:
    def __init__(self, transcript_id):
        self.transcript_id = transcript_id
        self.exons = []

    def sort(self):
        self.exons.sort(key=lambda e: (e.seqname, e.start, -e.end))

class Models(list):
    def __init__(self):
        self.by_transcript_id = {}

    def _obtain_transcript(self, transcript_id):
        trans = self.by_transcript_id.get(transcript_id)
        if trans is None:
            trans = self.by_transcript_id[transcript_id] = Transcript(transcript_id)
            self.append(trans)
        return trans

    def add_exon(self, exon):
        self._obtain_transcript(exon.transcript_id).exons.append(exon)

    def sort(self):
        for trans in self.by_transcript_id.values():
            trans.sort()

class GtfException(LrgaspException):
    pass

def fixup_empty_attr(gtf_df, col_name):
    """make empty None"""
    gtf_df[col_name] = np.where(gtf_df[col_name] == '', None, gtf_df[col_name])

def fixup_attr(gtf_df, col_name):
    """make an attribute None in empty or not specified"""
    if col_name in gtf_df.columns:
        fixup_empty_attr(gtf_df, col_name)
    else:
        gtf_df[col_name] = None

def fixup_gtf_attrs(gtf_df):
    fixup_attr(gtf_df, "transcript_id")
    fixup_attr(gtf_df, "gene_id")
    fixup_attr(gtf_df, "reference_gene_id")
    fixup_attr(gtf_df, "reference_transcript_id")

def load_exons(model_gtf):
    """load GTF exons into a list of Series objects of exons"""
    gtf_df = read_gtf(model_gtf)
    gtf_df = gtf_df.loc[gtf_df.feature == 'exon']
    if len(gtf_df) == 0:
        raise GtfException("no exon records found")
    fixup_gtf_attrs(gtf_df)
    return [gtf_df.iloc[i] for i in range(len(gtf_df))]

def rec_desc(rec):
    return "{} at {}:{}-{}".format(rec.feature, rec.seqname, rec.start, rec.end)

def validate_exon(exon):
    if exon.transcript_id is None:
        raise GtfException("must specify transcript_id attribute: " + rec_desc(exon))
    if exon.gene_id is None:
        raise GtfException("must specify gene_id attribute: " + rec_desc(exon))
    if exon.start > exon.end:
        raise GtfException("start must be <= end: " + rec_desc(exon))
    if exon.strand not in ('+', '-', '.'):
        raise GtfException("strand must be '+', '-', or '.': " + rec_desc(exon))

def validate_exons(exons):
    if len(exons) == 0:
        raise LrgaspException("no exons found in GTF file")
    for exon in exons:
        validate_exon(exon)

def build_transcripts(exons):
    models = Models()
    for exon in exons:
        models.add_exon(exon)
    return models

def check_trans_field_same(trans, attr):
    vals = set([e[attr] for e in trans.exons])
    if len(vals) > 1:
        raise GtfException("all exons in transcript {} must have same value for {}, found {}".format(trans.transcript_id, attr, list(sorted(vals))))

def validate_transcript(trans):
    # do gene_id first, as it can explain other problems
    check_trans_field_same(trans, "gene_id")
    check_trans_field_same(trans, "seqname")
    check_trans_field_same(trans, "strand")
    check_trans_field_same(trans, "reference_gene_id")
    check_trans_field_same(trans, "reference_transcript_id")

def validate_transcripts(models):
    if len(models.by_transcript_id) == 0:
        raise LrgaspException("no transcripts found in GTF file")
    for trans in models.by_transcript_id.values():
        validate_transcript(trans)

def load(model_gtf):
    """Validate GTF,returns exons grouped into transcripts"""
    try:
        exons = load_exons(model_gtf)
        validate_exons(exons)
        models = build_transcripts(exons)
        validate_transcripts(models)
        return models
    except (LrgaspException, ParsingError, ValueError) as ex:
        raise GtfException("Parse of GTF failed: {}".format(model_gtf)) from ex



"""
read-to-model map parser/validator
"""
import sys
import csv
from collections import defaultdict, namedtuple
from lrgasp import LrgaspException, gopen
from lrgasp.defs import validate_feature_ident

csv.field_size_limit(sys.maxsize)

# note: pandas was a bad fit for this due to multi-indexing

READ_ID = "read_id"
TRANSCRIPT_ID = "transcript_id"

class ReadModelPair(namedtuple("ReadModelPair", ("read_id", "transcript_id"))):
    """a read mapped to a model or None"""
    pass

class ReadModelMap(list):
    """read to model map container, adding lazy multiple indexes"""
    def __init__(self):
        self._read_id_idx = None
        self._transcript_id_idx = None

    def add(self, pair):
        self.append(pair)

    def _build_idx(self, col_name):
        idx = defaultdict(list)
        for rec in self:
            val = getattr(rec, col_name)
            if val is not None:
                idx[val] = rec
        idx.default_factory = None
        return idx

    def get_by_read_id(self, read_id):
        if self._read_id_idx is None:
            self._read_id_idx = self._build_idx("read_id")
        return self._read_id_idx.get(read_id)

    def get_by_transcript_id(self, transcript_id):
        if self._transcript_id_idx is None:
            self._transcript_id_idx = self._build_idx("transcript_id")
        return self._transcript_id_idx.get(transcript_id)

def _parse_header(reader):
    header = next(reader, None)
    if header is None:
        raise LrgaspException("Empty TSV file, requires a header line")
    try:
        read_id_col = header.index(READ_ID)
    except ValueError:
        raise LrgaspException(READ_ID + " column required")
    try:
        trans_id_col = header.index(TRANSCRIPT_ID)
    except ValueError:
        raise LrgaspException(TRANSCRIPT_ID + " column required")
    return read_id_col, trans_id_col, len(header)

def _parse_row(read_id, transcript_id):
    validate_feature_ident(read_id)
    validate_feature_ident(transcript_id)
    if transcript_id == '*':
        transcript_id = None
    return ReadModelPair(read_id, transcript_id)

def _tsv_reader(fh):
    # allows for extra columns, but all rows must have the same
    reader = csv.reader(fh, dialect=csv.excel_tab)
    read_id_col, trans_id_col, width = _parse_header(reader)
    for row in reader:
        if len(row) != width:
            raise LrgaspException(f"TSV row requires {width} columns found {len(row)}: {row}")
        yield _parse_row(row[read_id_col], row[trans_id_col])

def load(model_map_tsv):
    read_model_map = ReadModelMap()
    try:
        with gopen(model_map_tsv) as fh:
            for pair in _tsv_reader(fh):
                read_model_map.add(pair)
        if len(read_model_map) == 0:
            raise LrgaspException("TSV contains no data")
        return read_model_map
    except (LrgaspException, FileNotFoundError, csv.Error) as ex:
        raise LrgaspException("Parse of reads-to-models TSV failed: {}".format(model_map_tsv)) from ex



# Copyright 2006-2018 Mark Diekhans
# Copyright of original unknown (https://goodcode.io/articles/python-dict-object/)

# dictionary with keys as object files, based on:
# https://goodcode.io/articles/python-dict-object/
from collections import defaultdict

class ObjDict(dict):
    """Dict object where keys are field names.
    This is useful for JSON by doing:
       json.load(fh, object_pairs_hook=ObjDict)
    """
    __slots__ = ()

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)


class DefaultObjDict(defaultdict):
    """defaultdict-based object where keys are field names.
    This is useful for JSON by doing:
       json.load(fh, object_pairs_hook=DefaultObjDict)
    """
    __slots__ = ()

    def __init__(self, dtype):
        super().__init__(dtype)

    def __getattr__(self, name):
        return self[name]

    def __setattr__(self, name, value):
        self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)



"""
Experiment metadata parsing and validation.
"""
import os.path as osp
import json
from collections import namedtuple, defaultdict
from lrgasp import LrgaspException, gopen
from lrgasp.objDict import ObjDict
from lrgasp.defs import Repository, Species, ExperimentType, DataCategory, Platform, ExpressionUnits, validate_symbolic_ident, EXPERIMENT_JSON
from lrgasp.metadata_validate import Field, check_from_defs, validate_url
from lrgasp.data_sets import get_lrgasp_rna_seq_metadata

fld_experiment_id = Field("experiment_id", validator=validate_symbolic_ident)
fld_experiment_type = Field("experiment_type", ExperimentType)
fld_description = Field("description")
fld_species = Field("species", Species)
fld_data_category = Field("data_category", DataCategory)
fld_libraries = Field("libraries", list, element_dtype=str, validator=validate_symbolic_ident)
fld_extra_libraries = Field("extra_libraries", list, optional=True, element_dtype=dict)
fld_units = Field("units", ExpressionUnits, optional=True)
fld_software = Field("software", list, element_dtype=dict)
fld_notes = Field("notes", allow_empty=True, optional=True)

experiment_fields = (
    fld_experiment_id,
    fld_experiment_type,
    fld_species,
    fld_description,
    fld_data_category,
    fld_libraries,
    fld_extra_libraries,
    fld_units,
    fld_software,
    fld_notes,
)

fld_acc = Field("acc")
fld_repository = Field("repository", Repository)

extra_library_fields = (
    fld_acc,
    fld_repository,
    fld_notes,
)

fld_name = Field("name")
fld_version = Field("version")
fld_url = Field("url", validator=validate_url)
fld_config = Field("url", allow_empty=True, optional=True)

experiment_software_fields = (
    fld_name,
    fld_version,
    fld_url,
    fld_config,
    fld_notes,
)

valid_file_content = frozenset(["reads", "subreads", "nanopore_signal"])

class RunType(namedtuple("RunType",
                         ("sample", "library_prep", "platform"))):
    "attributes describing a run"

    def __str__(self):
        return str(self.sample) + '/' + str(self.library_prep) + '/' + str(self.platform)


def get_extra_libraries(experiment):
    """get the extra libraries field, or empty if not specified"""
    return experiment.get("extra_libraries", ())

def find_dups(lst):
    dups = []
    found = set()
    for v in lst:
        if (v in found) and (v not in dups):
            dups.append(v)
        found.add(v)
    return sorted(dups)

def _get_libraries_file_metadata(rna_seq_md, experiment):
    """collect metadata for all library files in experiment"""
    # exception if unknown acc
    return [rna_seq_md.get_file_by_acc(file_acc)
            for file_acc in experiment.libraries]

def library_validate(experiment, rna_seq_md, file_md):
    run_md = rna_seq_md.get_run_by_acc(file_md.run_acc)
    if run_md.species != experiment.species:
        raise LrgaspException(f"LRGASP RNA Seq library {run_md.run_acc} file {file_md.file_acc} is for species {run_md.species} while experiment {experiment.experiment_id} specifies species as {experiment.species}")

    # this should never happen, as the data matrix should be restrict to types we allow
    if file_md.output_type not in valid_file_content:
        raise LrgaspException(f"File {file_md.file_acc} of output_type '{file_md.output_type}' not support for LRGASP, "
                              "valid types are {}; please contact LRGASP project if this type of file is needed".format(", ".join(sorted(valid_file_content))))

def _fmt_run_types(run_types):
    "make set of RunType objects pretty"
    if len(run_types) == 0:
        return "none"
    else:
        return ", ".join([str(rt) for rt in sorted(run_types)])

def get_file_run_type(rna_seq_md, expr_file_md):
    run_md = rna_seq_md.get_run_by_acc(expr_file_md.run_acc)
    return RunType(run_md.sample, run_md.library_prep, run_md.platform)

def get_run_types(rna_seq_md, expr_file_mds):
    "group into sets of long and short RunType,"
    long_run_types = set()
    short_run_types = set()
    for file_md in expr_file_mds:
        run_md = rna_seq_md.get_run_by_acc(file_md.run_acc)
        rt = RunType(run_md.sample, run_md.library_prep, run_md.platform)
        if rt.platform == Platform.Illumina:
            short_run_types.add(rt)
        else:
            long_run_types.add(rt)
    return long_run_types, short_run_types

def get_run_type_descs(rna_seq_md, expr_file_mds):
    "produce string descriptions of file acc to run types for error messages"
    rt_to_file_acc = defaultdict(set)
    for expr_file_md in expr_file_mds:
        rt = get_file_run_type(rna_seq_md, expr_file_md)
        rt_to_file_acc[rt].add(expr_file_md.file_acc)

    return ["{}: {}".format(str(rt), ", ".join(sorted(rt_to_file_acc[rt])))
            for rt in sorted(rt_to_file_acc.keys())]


def libraries_validate_compat(experiment, rna_seq_md, expr_file_mds):
    """compatibility between libraries and experiments;"""
    data_category = experiment.data_category
    long_run_types, short_run_types = get_run_types(rna_seq_md, expr_file_mds)

    def _run_type_err_msg(rts):
        return "found: " + _fmt_run_types(rts) + " in these specified library files:\n    " + "\n    ".join(get_run_type_descs(rna_seq_md, expr_file_mds))

    if data_category == DataCategory.short_only:
        if len(long_run_types) > 0:
            raise LrgaspException(f"{data_category} experiments must not use long-read platforms, " + _run_type_err_msg(long_run_types))
        if len(short_run_types) != 1:
            raise LrgaspException(f"{data_category} experiments must use one and only one short read library/platform, " + _run_type_err_msg(short_run_types))
    elif data_category == DataCategory.long_only:
        if len(long_run_types) != 1:
            raise LrgaspException(f"{data_category} experiments must use one and only one long-read library/platform, " + _run_type_err_msg(long_run_types))
        if len(short_run_types) != 0:
            raise LrgaspException(f"{data_category} experiments must not use short read platforms, " + _run_type_err_msg(short_run_types))
    elif data_category == DataCategory.long_short:
        if len(long_run_types) != 1:
            raise LrgaspException(f"{data_category} experiments must use one and only one long-read library/platform, " + _run_type_err_msg(long_run_types))
        if len(short_run_types) != 1:
            raise LrgaspException(f"{data_category} experiments must use one and only one short read library/platform, " + _run_type_err_msg(short_run_types))
    elif data_category == DataCategory.kitchen_sink:
        if len(experiment.libraries) == 0:
            raise LrgaspException(f"{data_category} experiments must use some LRGASP RNA-Seq libraries")
    else:
        raise LrgaspException("bug")

def libraries_validate_paired_end(rna_seq_md, expr_file_mds):
    "make sure both paired ends are there"
    paired_expr_file_mds = [fm for fm in expr_file_mds if fm.paired_end is not None]
    paired_acc = frozenset([fm.file_acc for fm in paired_expr_file_mds])
    for file_md in paired_expr_file_mds:
        if file_md.paired_with not in paired_acc:
            raise LrgaspException(f"{file_md.file_acc} is a paired-end experiment, however the paired file {file_md.paired_with} is not specified in experiment.libraries")

def libraries_validate(experiment):
    rna_seq_md = get_lrgasp_rna_seq_metadata()
    dups = find_dups(experiment.libraries)
    if len(dups) > 0:
        raise LrgaspException(f"duplicate accession in libraries: {dups}")

    # per library validation
    expr_file_mds = _get_libraries_file_metadata(rna_seq_md, experiment)
    for file_md in expr_file_mds:
        library_validate(experiment, rna_seq_md, file_md)

    # cross-library validations
    libraries_validate_compat(experiment, rna_seq_md, expr_file_mds)
    libraries_validate_paired_end(rna_seq_md, expr_file_mds)

def extra_library_validate(extra_library):
    desc = "experiment.extra_libraries"
    check_from_defs(desc, extra_library_fields, extra_library)

def extra_libraries_validate(experiment):
    if experiment.data_category != DataCategory.kitchen_sink:
        raise LrgaspException("experiment extra_libraries may only be specified for kitchen_sink experiments")
    dups = find_dups([el.acc for el in experiment.extra_libraries])
    if len(dups) > 0:
        raise LrgaspException(f"duplicate accession in extra libraries: {dups}")
    for extra_library in experiment.extra_libraries:
        extra_library_validate(extra_library)

def experssion_units_validate(experiment):
    if experiment.experiment_type is ExperimentType.model:
        if fld_units.name in experiment:
            raise LrgaspException(f"experiment type {experiment.experiment_type} must not specify {fld_units.name}")
    else:
        if fld_units.name not in experiment:
            raise LrgaspException(f"experiment type {experiment.experiment_type} must specify {fld_units.name}")

def experiment_validate(experiment):
    desc = "experiment"
    check_from_defs(desc, experiment_fields, experiment)
    libraries_validate(experiment)
    if len(get_extra_libraries(experiment)) > 0:
        extra_libraries_validate(experiment)
    experssion_units_validate(experiment)

def load(experiment_json):
    """load and validate experiment metadata"""
    try:
        with gopen(experiment_json) as fh:
            experiment = json.load(fh, object_pairs_hook=ObjDict)
    except json.decoder.JSONDecodeError as ex:
        raise LrgaspException(f"parse of experiment metadata (JSON) failed: {experiment_json}") from ex
    try:
        experiment_validate(experiment)
    except LrgaspException as ex:
        raise LrgaspException(f"validation of experiment metadata failed: {experiment_json}") from ex
    return experiment

def load_from_entry(entry, experiment_id):
    """load and validation experiment metadata given a entry,
    add experiment_dir to metadata"""
    experiment_dir = osp.join(entry.entry_dir, experiment_id)
    experiment_json = osp.join(experiment_dir, EXPERIMENT_JSON)
    try:
        experiment = load(experiment_json)
    except (LrgaspException, FileNotFoundError, ValueError) as ex:
        raise LrgaspException(f"error parse metadata for entry {entry.entry_id}, experiment {experiment_id}: {experiment_json}") from ex
    experiment.experiment_dir = experiment_dir
    experiment.experiment_json = experiment_json
    return experiment



import os.path as osp
import sys
import gzip
import traceback

__version__ = "0.8.0"

class LrgaspException(Exception):
    pass

# exceptions that should result in a call to handle_prog_errors
prog_error_excepts = (LrgaspException, FileNotFoundError)

def handle_prog_errors(ex, debug):
    """Prints error messages without call stack and exit. For expected exceptions """

    print("Error: " + str(ex), file=sys.stderr)
    if debug:
        traceback.print_tb(ex.__traceback__, file=sys.stderr)
    exc = ex.__cause__
    while exc is not None:
        print("caused by: " + str(exc), file=sys.stderr)
        if debug:
            traceback.print_tb(exc.__traceback__, file=sys.stderr)
        exc = exc.__cause__
    exit(1)

def defined_file_path(dirname, uncomp_name):
    """get the path to a file that must exist and may optional be compressed"""
    p = osp.join(dirname, uncomp_name)
    if osp.exists(p):
        return p
    comp_name = uncomp_name + ".gz"
    p = osp.join(dirname, comp_name)
    if osp.exists(p):
        return p
    raise LrgaspException(f"can't find required file {comp_name} or {uncomp_name} in {dirname}")

def gopen(path):
    "open a file for reading, allowing compressed files"
    if path.endswith(".gz"):
        return gzip.open(path, "rt")
    else:
        return open(path)



"""
expression matrix parser/validator
"""
import csv
import pandas as pd
import numpy as np
from lrgasp import LrgaspException
from lrgasp import gopen
from lrgasp.defs import validate_feature_ident

def check_row_consistency(expression_tsv):
    """Pandas just pads or ignores inconsistent number of columns,
    so to validate the matrix is consistent, we need to make another pass.
    """
    minLen = maxLen = None
    with gopen(expression_tsv) as fh:
        for row in csv.reader(fh, dialect=csv.excel_tab):
            if minLen is None:
                minLen = maxLen = len(row)
            else:
                minLen = min(len(row), minLen)
                maxLen = max(len(row), maxLen)
    if minLen is None:
        raise LrgaspException("TSV is empty")
    if minLen != maxLen:
        raise LrgaspException(f"TSV has an inconsistent number of columns: min={minLen}, max={maxLen}")

def sample_column_names(expression):
    "return all but ID column"
    return [col for col in expression.columns if col != "ID"]

def validate_header(expression):
    if "ID" not in expression.columns:
        raise LrgaspException("TSV must have column 'ID'")
    if len(expression.columns) < 2:
        raise LrgaspException("TSV must have at least one sample column")

def check_column_type(expression, col):
    "validate column type is correct"
    if expression.dtypes[col] not in (np.float64, np.int64):
        raise LrgaspException(f"Invalid value(s) in column '{col}', must be a number or NA, appears to contain other types, induced type is ({expression.dtypes[col]})")

def validate_data(expression):
    if expression.size == 0:
        raise LrgaspException("TSV contains no data")
    expression.ID.apply(validate_feature_ident)
    for col in sample_column_names(expression):
        check_column_type(expression, col)

def load(expression_tsv):
    try:
        check_row_consistency(expression_tsv)
        with gopen(expression_tsv) as fh:
            expression = pd.read_csv(fh,
                                     dialect=csv.excel_tab,
                                     header=0,
                                     converters={"ID": str})
            validate_header(expression)
            expression.set_index("ID", drop=False, inplace=True, verify_integrity=True)
            validate_data(expression)
        return expression
    except (LrgaspException, pd.errors.ParserError, pd.errors.EmptyDataError, ValueError) as ex:
        raise LrgaspException("Parse of expression matrix TSV failed: {}".format(expression_tsv)) from ex



"""
Entry metadata parsing and validation.
"""
import os.path as osp
import json
from lrgasp import LrgaspException, gopen
from lrgasp.objDict import ObjDict
from lrgasp.metadata_validate import Field, check_from_defs, validate_email
from lrgasp.defs import Challenge, validate_symbolic_ident, validate_synapse_ident, validate_entry_ident, ENTRY_JSON

##
# top-level entry struct (from JSON)
##
fld_entry_id = Field("entry_id", validator=validate_entry_ident)
fld_challenge_id = Field("challenge_id", Challenge)
fld_team_id = Field("team_id", validator=validate_synapse_ident)
fld_team_name = Field("team_name")
fld_experiment_ids = Field("experiment_ids", list, element_dtype=str, validator=validate_symbolic_ident)
fld_notes = Field("notes", allow_empty=True, optional=True)
fld_contacts = Field("contacts", list, element_dtype=dict)

entry_fields = (
    fld_entry_id,
    fld_challenge_id,
    fld_team_id,
    fld_team_name,
    fld_experiment_ids,
    fld_notes,
    fld_contacts
)

##
# contacts entries
##
fld_contact_name = Field("name", str)
fld_contact_email = Field("email", str, validator=validate_email)

entry_contact_fields = (
    fld_contact_name,
    fld_contact_email,
    fld_notes
)

def entry_contact_validate(contact):
    desc = "entry.contacts"
    check_from_defs(desc, entry_contact_fields, contact)

def entry_validate(entry):
    desc = "entry"
    check_from_defs(desc, entry_fields, entry)
    try:
        validate_symbolic_ident(entry.entry_id)
    except LrgaspException as ex:
        raise LrgaspException(f"invalid {desc} {fld_entry_id.name}") from ex

    challenge_id = validate_entry_ident(entry.entry_id)
    if entry.challenge_id != challenge_id:
        raise LrgaspException(f"invalid {desc} {fld_entry_id.name} entry_id {entry.entry-id} prefix does not match challenge_id {entry.challenge_id}")
    for contact in entry.contacts:
        entry_contact_validate(contact)

def load(entry_json):
    """load and validate entry metadata"""
    try:
        with gopen(entry_json) as fh:
            entry = json.load(fh, object_pairs_hook=ObjDict)
    except json.decoder.JSONDecodeError as ex:
        raise LrgaspException(f"parse of entry metadata (JSON) failed: {entry_json}") from ex
    try:
        entry_validate(entry)
    except LrgaspException as ex:
        raise LrgaspException(f"validation of entry metadata failed: {entry_json}") from ex
    return entry

def load_dir(entry_dir):
    """load entry metadata, verifying that the file system directory matches
    the entry_id.  Save entry_dir in metadata """
    entry_json = osp.join(entry_dir, ENTRY_JSON)
    entry = load(entry_json)
    entry_base_dir = osp.basename(entry_dir)
    if entry_base_dir != entry.entry_id:
        raise LrgaspException(f"entry {entry.entry_id} must be in an directory {entry.entry_id}, not {entry_base_dir}")
    entry.entry_dir = entry_dir
    entry.entry_json = entry_json
    return entry
